package com.loafle.commons.rpc.util;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Type;
import java.util.HashMap;
import java.util.Map;

import com.loafle.commons.rpc.RPCException;
import com.loafle.commons.rpc.annotation.RPCMethod;

/**
 * RPCServiceRegistry
 */
public class RPCServiceRegistry {
  private Map<String, RPCServiceMeta> services;

  public RPCServiceRegistry() {
    this.services = new HashMap<>();
  }

  public void register(Object receiver, Class<?> receiverClass, String name) throws RPCException {
    if (null == receiver) {
      throw new RPCException("receiver is null");
    }

    String sName = name;
    if ("" == sName.trim()) {
      sName = receiverClass.getName();
    }

    RPCServiceMeta serviceMeta = new RPCServiceMeta();
    serviceMeta.instance = receiver;

    Method[] methods = receiverClass.getMethods();

    for (Method method : methods) {
      RPCMethod a = method.getAnnotation(RPCMethod.class);
      if (null == a) {
        continue;
      }

      RPCServiceMethodMeta methodMeta = new RPCServiceMethodMeta();
      methodMeta.method = method;
      methodMeta.returnType = method.getReturnType();
      methodMeta.parameterTypes = method.getGenericParameterTypes();

      String n = "" == a.value() ? method.getName() : a.value();
      serviceMeta.methods.put(n, methodMeta);
    }

    if (0 < serviceMeta.methods.size()) {
      this.services.put(sName, serviceMeta);
    }
  }

  public boolean hasMethod(String method) {
    String[] sm = method.split(".");

    if (null == sm || 2 != sm.length) {
      return false;
    }
    RPCServiceMeta serviceMeta = this.services.get(sm[0]);
    if (null == serviceMeta) {
      return false;
    }
    RPCServiceMethodMeta methodMeta = serviceMeta.getMethodMeta(sm[1]);
    if (null == methodMeta) {
      return false;
    }
    return true;
  }

  public Type[] getParameterTypes(String method) throws RPCException {
    String[] sm = method.split(".");

    if (null == sm || 2 != sm.length) {
      throw new RPCException(String.format("method[%s] is not valid", method));
    }
    RPCServiceMeta serviceMeta = this.services.get(sm[0]);
    if (null == serviceMeta) {
      throw new RPCException(String.format("service[%s] is not exist", sm[0]));
    }
    RPCServiceMethodMeta methodMeta = serviceMeta.getMethodMeta(sm[1]);
    if (null == methodMeta) {
      throw new RPCException(String.format("method[%s] of service[%s] is not exist", sm[1], sm[0]));
    }
    return methodMeta.parameterTypes;
  }

  public Object invoke(String method, Object... params) throws RPCException {
    String[] sm = method.split(".");

    if (null == sm || 2 != sm.length) {
      throw new RPCException(String.format("method[%s] is not valid", method));
    }
    RPCServiceMeta serviceMeta = this.services.get(sm[0]);
    if (null == serviceMeta) {
      throw new RPCException(String.format("service[%s] is not exist", sm[0]));
    }
    RPCServiceMethodMeta methodMeta = serviceMeta.getMethodMeta(sm[1]);
    if (null == methodMeta) {
      throw new RPCException(String.format("method[%s] of service[%s] is not exist", sm[1], sm[0]));
    }

    try {
      return methodMeta.method.invoke(serviceMeta.instance, params);
    } catch (IllegalAccessException | IllegalArgumentException | InvocationTargetException e) {
      throw new RPCException(String.format("Cannot invoke method[%s] with params[%s]", method, params), e);
    }
  }

  public static class RPCServiceMeta {
    private Object instance;
    private Map<String, RPCServiceMethodMeta> methods;

    RPCServiceMeta() {
      this.methods = new HashMap<>();
    }

    public Object getInstance() {
      return this.instance;
    }

    public RPCServiceMethodMeta getMethodMeta(String name) {
      return this.methods.get(name);
    }
  }

  public static class RPCServiceMethodMeta {
    private Method method;
    private Class<?> returnType;
    private Type[] parameterTypes;

    RPCServiceMethodMeta() {
    }

    public Method getMethod() {
      return this.method;
    }

    public Class<?> getReturnType() {
      return this.returnType;
    }

    public Type[] getParameterTypes() {
      return this.parameterTypes;
    }
  }
}